"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

__author__ = 'Fernando Serena'

import logging
from datetime import datetime as dt
import multiprocessing

from concurrent.futures.thread import ThreadPoolExecutor
from concurrent.futures import wait, ALL_COMPLETED
import networkx as nx

from agora.fountain.index import core as index, seeds

log = logging.getLogger('agora.fountain.paths')

pgraph = nx.DiGraph()
match_elm_cycles = {}


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    if n:
        for i in xrange(0, len(l), n):
            yield l[i:i + n]


def __build_directed_graph(generic=False, graph=None):
    """

    :return:
    """
    if graph is None:
        graph = nx.DiGraph()
    else:
        graph.clear()

    graph.add_nodes_from(index.get_types(), ty='type')
    for node in index.get_properties():
        p_dict = index.get_property(node)
        dom = set(p_dict.get('domain'))
        if generic:
            dom = filter(lambda x: not set.intersection(set(index.get_type(x)['super']), dom), dom)
        ran = set(p_dict.get('range'))
        if generic:
            try:
                ran = filter(lambda x: not set.intersection(set(index.get_type(x)['super']), ran), ran)
            except TypeError:
                pass
        edges = [(d, node) for d in dom]
        if p_dict.get('type') == 'object':
            edges.extend([(node, r) for r in ran])
        graph.add_edges_from(edges)
        graph.add_node(node, ty='prop', object=p_dict.get('type') == 'object', range=ran)

    print 'graph', list(graph.edges())
    return graph


def __build_paths(node, root, steps=None, level=0, path_graph=None, cache=None):
    """

    :param node:
    :param root:
    :param steps:
    :return:
    """

    def get_property(p):
        if p not in cache:
            cache[p] = index.get_property(p)
        return cache[p]

    def contains_cycle(graph):
        return bool(list(nx.simple_cycles(graph)))

    paths = []
    if steps is None:
        steps = []
    if path_graph is None:
        path_graph = nx.DiGraph()
    if cache is None:
        cache = {}

    log.debug('[{}] building paths to {}, with root {} and {} previous steps'.format(level, node, root, len(steps)))

    pred = set(pgraph.predecessors(node))
    for t in [x for x in pred]:
        new_path_graph = path_graph.copy()
        new_path_graph.add_nodes_from([t, node])
        new_path_graph.add_edges_from([(t, node)])
        if contains_cycle(new_path_graph):
            continue

        step = {'property': node, 'type': t}
        path = [step]
        new_steps = steps[:]
        new_steps.append(step)
        log.debug('[{}] added a new step {} in the path to {}'.format(level, (t, node), node))

        try:
            node_inverse = get_property(node)['inverse']
        except TypeError:
            node_inverse = []

        any_subpath = False
        next_steps = [x for x in pgraph.predecessors(t) if x != root and x not in node_inverse]
        for p in next_steps:
            log.debug('[{}] following {} as a pred property of {}'.format(level, p, t))
            extended_new_path_graph = new_path_graph.copy()
            extended_new_path_graph.add_node(p)
            extended_new_path_graph.add_edges_from([(p, t)])
            if contains_cycle(extended_new_path_graph):
                continue
            sub_paths = __build_paths(p, root, new_steps[:], level=level + 1, path_graph=extended_new_path_graph,
                                      cache=cache)

            any_subpath = any_subpath or len(sub_paths)
            for sp in sub_paths:
                paths.append(path + sp)
        if (len(next_steps) and not any_subpath) or not len(next_steps):
            paths.append(path)

    log.debug('[{}] returning {} paths to {}, with root {} and {} previous steps'.format(level, len(paths), node, root,
                                                                                         len(steps)))
    return paths


def calculate_paths():
    """

    :return:
    """
    def __find_matching_cycles(_elm):
        for j, c in enumerate(g_cycles):
            extended_elm = [_elm]
            if index.is_type(_elm):
                extended_elm.extend(index.get_type(_elm)["super"])

            if len([c for e in extended_elm if e in c]):
                yield j

    def __store_path(_i, _path):
        pipe.zadd('paths:{}'.format(elm), _i, _path)

    def __calculate_node_paths(n, d):
        log.debug('[START] Calculating paths to {} with data {}'.format(n, d))
        _paths = []
        if d.get('ty') == 'type':
            for p in pgraph.predecessors(n):
                log.debug('Following root [{}] predecessor property {}'.format(n, p))
                _paths.extend(__build_paths(p, n))
        else:
            _paths.extend(__build_paths(n, n))
        log.debug('[END] {} paths for {}'.format(len(_paths), n))
        if len(_paths):
            node_paths[n] = _paths

    log.info('Calculating paths...')
    match_elm_cycles.clear()
    start_time = dt.now()

    __build_directed_graph(graph=pgraph)
    g_graph = __build_directed_graph(generic=True)

    cycle_keys = index.r.keys('*cycles*')
    for ck in cycle_keys:
        index.r.delete(ck)
    g_cycles = list(nx.simple_cycles(g_graph))
    with index.r.pipeline() as pipe:
        pipe.multi()
        for i, cy in enumerate(g_cycles):
            print cy
            cycle = []
            t_cycle = None
            for elm in cy:
                if index.is_type(elm):
                    t_cycle = elm
                elif t_cycle is not None:
                    cycle.append({'property': elm, 'type': t_cycle})
                    t_cycle = None
            if t_cycle is not None:
                cycle.append({'property': cy[0], 'type': t_cycle})
            pipe.zadd('cycles', i, cycle)
        pipe.execute()

    locks = __lock_key_pattern('paths:*')
    keys = [k for (k, _) in locks]
    if len(keys):
        index.r.delete(*keys)

    node_paths = {}
    futures = []
    with ThreadPoolExecutor(multiprocessing.cpu_count()) as th_pool:
        for node, data in pgraph.nodes(data=True):
            futures.append(th_pool.submit(__calculate_node_paths, node, data))
        wait(futures, timeout=None, return_when=ALL_COMPLETED)
        th_pool.shutdown()

    for ty in [_ for _ in index.get_types() if _ in node_paths]:
        for sty in [_ for _ in index.get_type(ty)['sub'] if _ in node_paths]:
            node_paths[ty].extend(node_paths[sty])

    node_paths = node_paths.items()

    print 'preparing to persist the calculated paths...{}'.format(len(node_paths))

    with index.r.pipeline() as pipe:
        pipe.multi()
        with ThreadPoolExecutor(multiprocessing.cpu_count()) as th_pool:
            for (elm, paths) in node_paths:
                futures = []
                for (i, path) in enumerate(paths):
                    futures.append(th_pool.submit(__store_path, i, path))
                    for step in path:
                        step_ty = step.get('type')
                        if step_ty not in match_elm_cycles:
                            match_elm_cycles[step_ty] = __find_matching_cycles(step_ty)
                        step_pr = step.get('property')
                        if step_pr not in match_elm_cycles:
                            match_elm_cycles[step_pr] = __find_matching_cycles(step_pr)
                wait(futures, timeout=None, return_when=ALL_COMPLETED)
                pipe.execute()
            th_pool.shutdown()

        # Store type and property cycles
        for elm in match_elm_cycles.keys():
            for c in match_elm_cycles[elm]:
                pipe.sadd('cycles:{}'.format(elm), c)
        pipe.execute()
        for t in [_ for _ in index.get_types() if _ not in match_elm_cycles]:
            for c in __find_matching_cycles(t):
                pipe.sadd('cycles:{}'.format(t), c)
        pipe.execute()

    for _, l in locks:
        l.release()

    log.info('Found {} paths in {}ms'.format(len(index.r.keys('paths:*')),
                                             (dt.now() - start_time).total_seconds() * 1000))


def __lock_key_pattern(pattern):
    """

    :param pattern:
    :return:
    """
    pattern_keys = index.r.keys(pattern)
    for k in pattern_keys:
        yield k, index.r.lock(k)


def __detect_and_remove_cycles(cycle, steps):
    """

    :param cycle:
    :param steps:
    :return:
    """
    if cycle[0] in steps:
        steps_copy = steps[:]
        start_index = steps_copy.index(cycle[0])
        end_index = start_index + len(cycle)
        try:
            cand_cycle = steps_copy[start_index:end_index]
            if end_index >= len(steps_copy):
                cand_cycle.extend(steps_copy[:end_index - len(steps_copy)])
            if cand_cycle == cycle:
                steps_copy = steps[0:start_index - end_index + len(steps_copy)]
                if len(steps) > end_index:
                    steps_copy += steps[end_index:]
        except IndexError:
            pass
        return steps_copy
    return steps


def find_path(elm):
    """

    :param elm:
    :return:
    """
    seed_av = {}

    def check_seed_availability(ty):
        if ty not in seed_av:
            seed_av[ty] = seeds.get_type_seeds(ty)
        return seed_av[ty]

    def filter_path_cycles(_seeds):
        """

        :param _seeds:
        :return:
        """
        sub_steps = list(reversed(path[:step_index + 1]))
        for _step in sub_steps:
            cycle_ids.update([int(c) for c in index.r.smembers('cycles:{}'.format(_step.get('type')))])
        cycles = [eval(index.r.zrangebyscore('cycles', c, c).pop()) for c in cycle_ids]
        sub_path = {'cycles': list(cycle_ids), 'seeds': _seeds, 'steps': sub_steps}

        # for cycle in sorted(cycles, key=lambda x: len(x), reverse=True):  # First filter bigger cycles
        #     sub_steps = __detect_and_remove_cycles(cycle, sub_steps)
        sub_path['steps'] = sub_steps
        if sub_path not in seed_paths:
            seed_paths.append(sub_path)
        return cycle_ids

    seed_paths = []
    paths = [(int(score), eval(path)) for path, score in index.r.zrange('paths:{}'.format(elm), 0, -1, withscores=True)]

    applying_cycles = set([])
    cycle_ids = set([int(c) for c in index.r.smembers('cycles:{}'.format(elm))])

    step_index = 0
    for score, path in paths:
        for step_index, step in enumerate(path):
            ty = step.get('type')
            type_seeds = check_seed_availability(ty)
            if len(type_seeds):
                seed_cycles = filter_path_cycles(type_seeds)
                applying_cycles = applying_cycles.union(set(seed_cycles))

    # It only returns seeds if elm is a type and there are seeds of it
    req_type_seeds = check_seed_availability(elm)
    if len(req_type_seeds):
        path = []
        seed_cycles = filter_path_cycles(req_type_seeds)
        applying_cycles = applying_cycles.union(set(seed_cycles))

    applying_cycles = [{'cycle': int(cid), 'steps': eval(index.r.zrange('cycles', cid, cid).pop())} for cid in
                       applying_cycles]
    return list(seed_paths), applying_cycles


# Build the current graph on import
__build_directed_graph(graph=pgraph)



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

__author__ = 'Fernando Serena'

import unittest
import json
from urlparse import urlparse
import logging

from nose.tools import *

from agora.fountain.server import app
from agora.fountain.tests.util import AgoraGraph


def setup():
    from agora.fountain.server.config import TestingConfig

    app.config['TESTING'] = True
    app.config.from_object(TestingConfig)
    app.config['STORE'] = 'memory'

    from agora.fountain.index.core import r
    r.flushdb()

    from agora.fountain import api


class FountainTest(unittest.TestCase):
    @classmethod
    def tearDownClass(cls):
        from agora.fountain.index.core import r
        r.flushdb()

        test_client = app.test_client()
        vocabs = json.loads(test_client.get('/vocabs').data)
        for v_uri in vocabs.values():
            rv = test_client.delete(v_uri)
            eq_(rv.status_code, 200, "The resource couldn't be deleted")

    @classmethod
    def setUpClass(cls):
        from agora.fountain.index.core import r
        r.flushdb()

    def setUp(self):
        self.app = app.test_client()
        self._graph = None
        self.log = logging.getLogger('agora.fountain.test')

    def get_vocabularies(self):
        return json.loads(self.get('/vocabs'))

    def get_vocabulary(self, uri):
        path = urlparse(uri).path
        return self.get(path)

    def post_vocabulary(self, filename, exp_code=201):
        with open('agora/fountain/tests/vocabs/{}.ttl'.format(filename)) as f:
            vocab = f.read()
            response = self.post('/vocabs', vocab, message='The vocabulary was not created properly', exp_code=exp_code)
            if exp_code != 201:
                return None
            return response.headers.get('Location', None)

    def delete_vocabulary(self, uri):
        path = urlparse(uri).path
        self.delete(path, 'The test vocabulary should exist previously')

    def post_seed(self, ty, uri, exp_code=201):
        response = self.post('/seeds', json.dumps({"uri": uri, "type": ty}),
                             content_type='application/json', exp_code=exp_code)

        return response.headers.get('Location', None)

    def delete_seed(self, uri):
        path = urlparse(uri).path
        self.delete(path, 'That seed should exist previously')

    def get_type_seeds(self, ty):
        return json.loads(self.get('/seeds/{}'.format(ty)))["seeds"]

    @property
    def seeds(self):
        return json.loads(self.get('/seeds'))["seeds"]

    def get_paths(self, node):
        paths = json.loads(self.get('paths/{}'.format(node)))
        return paths["paths"], paths["all-cycles"]

    def get(self, path, exp_code=200, error_message=None):
        rv = self.app.get(path)
        if error_message is None:
            error_message = 'There is a problem with the request'
        eq_(rv.status_code, exp_code, error_message)
        return rv.data

    def post(self, path, data, content_type='text/turtle', exp_code=201, message=None):
        rv = self.app.post(path, data=data, headers={'Content-Type': content_type})
        if message is None:
            message = 'The resource was not created properly'
        eq_(rv.status_code, exp_code, message + ": %s" % rv.status_code)
        return rv

    def delete(self, path, error_message=None):
        rv = self.app.delete(path)
        if error_message is None:
            error_message = "The resource couldn't be deleted"
        eq_(rv.status_code, 200, error_message)
        return rv.data

    @property
    def types(self):
        return json.loads(self.get('/types'))['types']

    @property
    def properties(self):
        return json.loads(self.get('/properties'))['properties']

    def get_type(self, ty):
        return json.loads(self.get('/types/{}'.format(ty)))

    def get_property(self, ty):
        return json.loads(self.get('/properties/{}'.format(ty)))

    @property
    def graph(self):
        if self._graph is None:
            _graph = AgoraGraph()
            types = self.types
            _graph.add_types_from(types)
            for node in self.properties:
                p_dict = self.get_property(node)
                dom = p_dict.get('domain')
                ran = p_dict.get('range')
                edges = [(d, node) for d in dom]
                if p_dict.get('type') == 'object':
                    edges.extend([(node, r) for r in ran])
                _graph.add_edges_from(edges)
                _graph.add_property(node, obj=p_dict.get('type') == 'object')
            for node in types:
                p_dict = self.get_type(node)
                refs = p_dict.get('refs')
                props = p_dict.get('properties')
                edges = [(r, node) for r in refs]
                edges.extend([(node, p) for p in props])
                _graph.add_edges_from(edges)
                if p_dict['super']:
                    _graph[node]['super'] = p_dict['super']
                if p_dict['sub']:
                    _graph[node]['sub'] = p_dict['sub']

        return _graph

    def check_property(self, name, domain=None, range=None, inverse=None):
        def check_edge(p_name, direction, func, expected):
            actual = func(p_name)
            if type(expected) is list:
                eq_(len(actual), len(expected), '{} must have {} {} type'.format(p_name, len(expected), direction))
                eq_(len(set.difference(set(actual), set(expected))), 0,
                    'Found wrong %s types: %s' % (direction, actual))
            elif len(actual):
                assert 'No %s was expected!' % direction

        check_edge(name, 'domain', self.graph.get_property_domain, domain)
        check_edge(name, 'range', self.graph.get_property_range, range)

        actual_inverse = self.graph.get_inverse_property(name)
        eq_(actual_inverse, inverse, 'Expected {} as inverse, but found: {}'.format(inverse, actual_inverse))

    def check_type(self, name, properties=None, refs=None):
        def check_attribute(p_name, attribute, func, expected):
            actual = func(p_name)
            if type(expected) is list:
                eq_(len(actual), len(expected), '{} must have {} {}'.format(p_name, len(expected), attribute))
                eq_(len(set.difference(set(actual), set(expected))), 0,
                    'Found wrong %s: %s' % (attribute, actual))
            elif len(actual):
                assert 'No %s was expected!' % attribute

        check_attribute(name, 'properties', self.graph.get_type_properties, properties)
        check_attribute(name, 'references', self.graph.get_type_refs, refs)



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at 

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

__author__ = 'Fernando Serena'

import base64

from agora.fountain.index.core import r
from agora.fountain.index import core as index
from agora.fountain.exceptions import FountainError


class TypeNotAvailableError(FountainError):
    pass


class DuplicateSeedError(FountainError):
    pass


class InvalidSeedError(FountainError):
    pass


def add_seed(uri, ty):
    from rfc3987 import parse
    parse(uri, rule='URI')
    type_found = False
    type_keys = r.keys('*:types')
    for tk in type_keys:
        if r.sismember(tk, ty):
            type_found = True
            encoded_uri = base64.b64encode(uri)
            if r.sismember('seeds:{}'.format(ty), encoded_uri):
                raise DuplicateSeedError('{} is already registered as a seed of type {}'.format(uri, ty))
            r.sadd('seeds:{}'.format(ty), base64.b64encode(uri))
            break

    if not type_found:
        raise TypeNotAvailableError("{} is not a valid type".format(ty))

    return base64.b64encode('{}|{}'.format(ty, uri))


def get_seed(sid):
    try:
        ty, uri = base64.b64decode(sid).split('|')
        if r.sismember('seeds:{}'.format(ty), base64.b64encode(uri)):
            return {'type': ty, 'uri': uri}
    except TypeError as e:
        raise InvalidSeedError(e.message)

    raise InvalidSeedError(sid)


def delete_seed(sid):
    try:
        ty, uri = base64.b64decode(sid).split('|')
        set_key = 'seeds:{}'.format(ty)
        encoded_uri = base64.b64encode(uri)
        if not r.srem(set_key, encoded_uri):
            raise InvalidSeedError(sid)
    except TypeError as e:
        raise InvalidSeedError(e.message)


def delete_type_seeds(ty):
    r.delete('seeds:{}'.format(ty))


def get_seeds():
    def iterator():
        seed_types = r.keys('seeds:*')
        for st in seed_types:
            ty = st.replace('seeds:', '')
            for seed in list(r.smembers(st)):
                yield ty, base64.b64decode(seed)

    import collections

    result_dict = collections.defaultdict(list)
    for t, uri in iterator():
        result_dict[t].append(uri)
    return result_dict


def get_type_seeds(ty):
    try:
        t_dict = index.get_type(ty)
        all_seeds = set([])
        for t in t_dict['sub'] + [ty]:
            all_seeds.update([base64.b64decode(seed) for seed in list(r.smembers('seeds:{}'.format(t)))])
        return list(all_seeds)
    except TypeError:
        # Check if it is a property...and return an empty list
        try:
            index.get_property(ty)
            return []
        except TypeError:
            raise TypeNotAvailableError(ty)



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

__author__ = 'Fernando Serena'

import json

from flask import make_response, request, jsonify, render_template, url_for
from flask_negotiate import consumes

from agora.fountain.vocab.schema import prefixes
import agora.fountain.index.core as index
import agora.fountain.index.seeds as seeds
from agora.fountain.view.path import view_path
from agora.fountain.view.graph import view_graph
from agora.fountain.index.paths import calculate_paths, find_path
import agora.fountain.vocab.onto as vocs
from agora.fountain.server import app


class APIError(Exception):
    status_code = 400

    def __init__(self, message, status_code=None, payload=None):
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['message'] = self.message
        return rv


class NotFound(APIError):
    def __init__(self, message, payload=None):
        super(NotFound, self).__init__(message, 404, payload)


class Conflict(APIError):
    def __init__(self, message, payload=None):
        super(Conflict, self).__init__(message, 409, payload)


@app.errorhandler(APIError)
def handle_invalid_usage(error):
    response = jsonify(error.to_dict())
    response.status_code = error.status_code
    return response


@app.route('/vocabs')
def get_vocabularies():
    """
    Return the currently used ontology
    :return:
    """
    vocabs = vocs.get_vocabularies()
    vocabs = [(x, url_for('get_vocabulary', vid=x, _external=True)) for x in vocabs]
    response = make_response(json.dumps(dict(vocabs)))
    response.headers['Content-Type'] = 'application/json'

    return response


@app.route('/vocabs/<vid>')
def get_vocabulary(vid):
    """
    Return a concrete vocabulary
    :param vid: The identifier of a vocabulary (prefix)
    :return:
    """
    response = make_response(vocs.get_vocabulary(vid))
    response.headers['Content-Type'] = 'text/turtle'

    return response


def __analyse_vocabularies(vids):
    for vid in reversed(vids):
        index.extract_vocabulary(vid)
    calculate_paths()


def __check_seeds():
    seed_dict = seeds.get_seeds()
    types = index.get_types()
    obsolete_types = set.difference(set(seed_dict.keys()), set(types))
    for t in obsolete_types:
        seeds.delete_type_seeds(t)


@app.route('/vocabs', methods=['POST'])
@consumes('text/turtle')
def add_vocabulary():
    """
    Add a new vocabulary to the fountain
    :return:
    """
    try:
        vids = vocs.add_vocabulary(request.data)
    except vocs.VocabularyNotFound, e:
        raise APIError('Ontology URI not found')
    except vocs.DuplicateVocabulary, e:
        raise Conflict(e.message)

    __analyse_vocabularies(vids)

    response = make_response()
    response.status_code = 201
    response.headers['Location'] = url_for('get_vocabulary', vid=vids.pop(0), _external=True)
    return response


@app.route('/vocabs/<vid>', methods=['PUT'])
@consumes('text/turtle')
def update_vocabulary(vid):
    """
    Updates an already contained vocabulary
    :return:
    """
    try:
        vocs.update_vocabulary(vid, request.data)
    except IndexError:
        raise APIError('Ontology URI not found')
    except vocs.UnknownVocabulary, e:
        raise NotFound(e.message)
    except Exception, e:
        raise APIError(e.message)

    __analyse_vocabularies([vid])
    __check_seeds()

    response = make_response()
    response.status_code = 200
    return response


@app.route('/vocabs/<vid>', methods=['DELETE'])
def delete_vocabulary(vid):
    """
    Delete an existing vocabulary
    :return:
    """
    try:
        vocs.delete_vocabulary(vid)
    except IndexError:
        raise APIError('Ontology URI not found')
    except vocs.UnknownVocabulary, e:
        raise NotFound(e.message)

    __analyse_vocabularies([vid])
    __check_seeds()

    response = make_response()
    response.status_code = 200
    return response


@app.route('/prefixes')
def get_prefixes():
    """
    Return the prefixes dictionary of the ontology
    :return:
    """
    return jsonify(prefixes())


@app.route('/types')
def get_types():
    """
    Return the list of supported types (prefixed)
    :return:
    """
    return jsonify({"types": index.get_types()})


@app.route('/types/<string:t>')
def get_type(t):
    """
    Return a concrete type description
    :param t: prefixed type e.g. foaf:Person
    :return: description of 't'
    """
    return jsonify(index.get_type(t))


@app.route('/properties')
def get_properties():
    """
    Return the list of supported properties (prefixed)
    :return:
    """
    return jsonify({"properties": index.get_properties()})


@app.route('/properties/<string:prop>')
def get_property(prop):
    """
    Return a concrete property description
    :param prop: prefixed property e.g. foaf:name
    :return: description of 'prop'
    """
    p = index.get_property(prop)

    return jsonify(p)


@app.route('/seeds')
def get_seeds():
    """
    Return the complete list of seeds available
    :return:
    """
    return jsonify({"seeds": seeds.get_seeds()})


@app.route('/seeds/<string:ty>')
def get_type_seeds(ty):
    """
    Return the list of seeds of a certain type
    :param ty: prefixed required type e.g. foaf:Person
    :return:
    """
    try:
        return jsonify({"seeds": seeds.get_type_seeds(ty)})
    except seeds.TypeNotAvailableError as e:
        raise NotFound(e.message)


@app.route('/seeds', methods=['POST'])
@consumes('application/json')
def add_seed():
    """
    Add a new seed of a specific supported type
    :return:
    """
    data = request.json
    try:
        sid = seeds.add_seed(data.get('uri', None), data.get('type', None))
        response = make_response(sid)
        response.headers['Location'] = url_for('get_seed', sid=sid, _external=True)
        response.status_code = 201
    except (seeds.TypeNotAvailableError, ValueError) as e:
        raise APIError(e.message)
    except seeds.DuplicateSeedError as e:
        raise Conflict(e.message)
    return response


@app.route('/seeds/id/<sid>')
def get_seed(sid):
    """
    Get the known information about a specific seed by id
    :return:
    """
    try:
        seed = seeds.get_seed(sid)
        return jsonify(seed)
    except seeds.InvalidSeedError, e:
        raise NotFound(e.message)


@app.route('/seeds/id/<sid>', methods=['DELETE'])
def delete_seed(sid):
    """
    Delete a specific seed by id
    :return:
    """
    try:
        seeds.delete_seed(sid)
        return make_response()
    except seeds.InvalidSeedError, e:
        raise NotFound(e.message)


@app.route('/paths/<elm>')
@app.route('/paths/<elm>/view')
def get_path(elm):
    """
    Return a path to a specific elem (either a property or a type, always prefixed)
    :param elm: The required prefixed type/property
    :return:
    """
    try:
        seed_paths, all_cycles = find_path(elm)
        if 'view' in request.url_rule.rule:
            nodes, edges, roots = view_path(elm, seed_paths)
            return render_template('graph-path.html',
                                   nodes=json.dumps(nodes),
                                   edges=json.dumps(edges), roots=json.dumps(roots))
        else:
            return jsonify({'paths': seed_paths, 'all-cycles': all_cycles})
    except seeds.TypeNotAvailableError, e:
        raise APIError(e.message)


@app.route('/graph/')
def show_graph():
    nodes, edges, roots = view_graph()

    return render_template('graph-vocabs.html',
                           nodes=json.dumps(nodes),
                           edges=json.dumps(edges), roots=json.dumps(roots))



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at 

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
from rdflib.plugins.parsers.notation3 import BadSyntax

__author__ = 'Fernando Serena'

import StringIO
import urlparse

from rdflib import Graph, RDF
from rdflib.namespace import OWL

import agora.fountain.vocab.schema as sch


class VocabularyException(Exception):
    pass


class DuplicateVocabulary(VocabularyException):
    pass


class VocabularyNotFound(VocabularyException):
    pass


class UnknownVocabulary(VocabularyException):
    pass


def __load_owl(owl):
    """

    :param owl:
    :return:
    """
    owl_g = Graph()
    owl_g.parse(source=StringIO.StringIO(owl), format='turtle')

    try:
        uri = list(owl_g.subjects(RDF.type, OWL.Ontology)).pop()
        vid = [p for (p, u) in owl_g.namespaces() if uri in u and p != '']
        imports = owl_g.objects(uri, OWL.imports)
        if not len(vid):
            vid = urlparse.urlparse(uri).path.split('/')[-1]
        else:
            vid = vid.pop()

        return vid, uri, owl_g, imports
    except IndexError:
        raise VocabularyNotFound()


def add_vocabulary(owl):
    """

    :param owl:
    :return:
    """
    vid, uri, owl_g, imports = __load_owl(owl)

    if vid in sch.contexts():
        raise DuplicateVocabulary('Vocabulary already contained')

    sch.add_context(vid, owl_g)
    vids = [vid]

    # TODO: Import referenced ontologies
    for im_uri in imports:
        print im_uri
        im_g = Graph()
        try:
            im_g.load(im_uri, format='turtle')
        except BadSyntax:
            try:
                im_g.load(im_uri)
            except BadSyntax:
                print 'bad syntax in {}'.format(im_uri)

        try:
            child_vids = add_vocabulary(im_g.serialize(format='turtle'))
            vids.extend(child_vids)
        except DuplicateVocabulary, e:
            print 'already added'
        except VocabularyNotFound, e:
            print 'uri not found for {}'.format(im_uri)
        except Exception, e:
            print e.message

    return vids


def update_vocabulary(vid, owl):
    """

    :param vid:
    :param owl:
    :return:
    """
    owl_vid, uri, owl_g, imports = __load_owl(owl)

    if vid != owl_vid:
        raise Exception("Identifiers don't match")

    if vid not in sch.contexts():
        raise UnknownVocabulary('Vocabulary id is not known')

    sch.update_context(vid, owl_g)


def delete_vocabulary(vid):
    """

    :param vid:
    :return:
    """
    if vid not in sch.contexts():
        raise UnknownVocabulary('Vocabulary id is not known')

    sch.remove_context(vid)


def get_vocabularies():
    """

    :return:
    """
    return sch.contexts()


def get_vocabulary(vid):
    """

    :param vid:
    :return:
    """
    return sch.get_context(vid).serialize(format='turtle')



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

__author__ = 'Fernando Serena'
import logging

from rdflib import ConjunctiveGraph, URIRef, BNode
from rdflib.namespace import RDFS

from agora.fountain.server import app

log = logging.getLogger('agora_fountain.schema')

store_mode = app.config['STORE']
if 'persist' in store_mode:
    graph = ConjunctiveGraph('Sleepycat')
    graph.open('graph_store', create=True)
else:
    graph = ConjunctiveGraph()

log.info('Loading ontology...'),
graph.store.graph_aware = False
log.debug('\n{}'.format(graph.serialize(format='turtle')))
log.info('Ready')

_namespaces = {}
_prefixes = {}


def __flat_slice(lst):
    """

    :param lst:
    :return:
    """
    lst = list(lst)
    for i, _ in enumerate(lst):
        while hasattr(lst[i], "__iter__") and not isinstance(lst[i], basestring):
            lst[i:i + 1] = lst[i]
    return set(filter(lambda x: x is not None, lst))


def __q_name(uri):
    """

    :param uri:
    :return:
    """
    q = uri.n3(graph.namespace_manager)
    return q


def __extend_prefixed(pu):
    """

    :param pu:
    :return:
    """
    parts = pu.split(':')
    if len(parts) == 1:
        parts = ('', parts[0])
    try:
        return URIRef(_prefixes[parts[0]] + parts[1])
    except KeyError:
        return BNode(pu)


def prefixes(vid=None):
    """

    :param vid:
    :return:
    """
    context = graph
    if vid is not None:
        context = context.get_context(vid)
    return list(context.namespaces())


def contexts():
    """

    :return:
    """
    return [str(x.identifier) for x in graph.contexts()]


def update_context(vid, g):
    """

    :param vid:
    :param g:
    :return:
    """
    context = graph.get_context(vid)
    graph.remove_context(context)
    add_context(vid, g)


def remove_context(vid):
    """

    :param vid:
    :return:
    """
    context = graph.get_context(vid)
    graph.remove_context(context)


def get_context(vid):
    """

    :param vid:
    :return:
    """
    return graph.get_context(vid)


def add_context(vid, g):
    """

    :param vid:
    :param g:
    :return:
    """
    vid_context = graph.get_context(vid)
    for t in g.triples((None, None, None)):
        vid_context.add(t)

    for (p, u) in g.namespaces():
        if p != '':
            vid_context.bind(p, u)

    _namespaces.update([(uri, prefix) for (prefix, uri) in graph.namespaces()])
    _prefixes.update([(prefix, uri) for (prefix, uri) in graph.namespaces()])


def get_types(vid=None):
    """

    :param vid:
    :return:
    """
    context = graph
    if vid is not None:
        context = context.get_context(vid)
    types = set([])
    q_class_result = context.query(
        """SELECT DISTINCT ?c ?x WHERE {
             {
               {?c a owl:Class}
               UNION
               {?c a rdfs:Class}
               FILTER(isURI(?c))
             }
             UNION
             {?c rdfs:subClassOf ?x FILTER(isURI(?x))}
           }""")
    classes_set = __flat_slice(q_class_result)
    types.update([__q_name(c) for c in classes_set])
    q_class_result = context.query(
        """SELECT DISTINCT ?r ?d WHERE {
             ?p a owl:ObjectProperty.
             {?p rdfs:range ?r FILTER(isURI(?r))}
             UNION
             {?p rdfs:domain ?d FILTER(isURI(?r))}
           }""")
    classes_set = __flat_slice(q_class_result)
    types.update([__q_name(c) for c in classes_set])
    types.update([__q_name(x[0]) for x in context.query(
        """SELECT DISTINCT ?c WHERE {
             {?r owl:allValuesFrom ?c}
             UNION
             {?a owl:someValuesFrom ?c}
             UNION
             {?b owl:onClass ?c}
             FILTER (isURI(?c))
           }""")])

    return types


def get_properties(vid=None):
    """

    :param vid:
    :return:
    """
    context = graph
    if vid is not None:
        context = context.get_context(vid)
    properties = set([])
    properties.update([__q_name(o or d) for (o, d) in
                       context.query(
                           """SELECT DISTINCT ?o ?d WHERE {
                                {?d a rdf:Property}
                                UNION
                                {?o a owl:ObjectProperty}
                                UNION
                                {?d a owl:DatatypeProperty}
                                UNION
                                {?r a owl:Restriction . ?r owl:onProperty ?o}
                              }""")])
    return properties


def get_property_domain(prop, vid=None):
    """

    :param prop:
    :param vid:
    :return:
    """
    context = graph
    if vid is not None:
        context = context.get_context(vid)
    dom = set([])
    domain_set = set([__q_name(c[0]) for c in context.query(
        """SELECT DISTINCT ?c WHERE {
             { %s rdfs:domain ?c }
             UNION
             { ?c rdfs:subClassOf [ owl:onProperty %s ] }
             FILTER (isURI(?c))
           }""" % (prop, prop))])
    dom.update(domain_set)
    for t in domain_set:
        dom.update(get_subtypes(t, vid))
    return dom


def is_object_property(prop, vid=None):
    """

    :param prop:
    :param vid:
    :return:
    """
    context = graph
    if vid is not None:
        context = context.get_context(vid)
    type_res = context.query("""ASK {%s a owl:ObjectProperty}""" % prop)
    object_evidence = [_ for _ in type_res].pop()

    if not object_evidence:
        object_evidence = [_ for _ in
                           context.query("""ASK {?r owl:onProperty %s.
                                    {?r owl:someValuesFrom ?o} UNION
                                    {?r owl:allValuesFrom ?a} UNION
                                    {?r owl:onClass ?c} .}""" % prop)].pop()
    return object_evidence


def get_property_range(prop, vid=None):
    """

    :param prop:
    :param vid:
    :return:
    """
    context = graph
    if vid is not None:
        context = context.get_context(vid)
    rang = set([])

    range_set = set([__q_name(r[0]) for r in context.query(
        """SELECT DISTINCT ?r WHERE {
             {%s rdfs:range ?r}
             UNION
             {?d owl:onProperty %s. ?d owl:allValuesFrom ?r. FILTER(isURI(?r))}
             UNION
             {?d owl:onProperty %s. ?d owl:someValuesFrom ?r. FILTER(isURI(?r))}
             UNION
             {?d owl:onProperty %s. ?d owl:onClass ?r. FILTER(isURI(?r))}
             UNION
             {?d owl:onProperty %s. ?d owl:onDataRange ?r. FILTER(isURI(?r))}
           }""" % (prop, prop, prop, prop, prop))])

    rang.update(range_set)
    for y in range_set:
        rang.update(get_subtypes(y, vid))
    return rang


def get_property_inverses(prop, vid=None):
    """

    :param prop:
    :param vid:
    :return:
    """
    context = graph
    if vid is not None:
        context = context.get_context(vid)
    inverses = set([])
    inverses.update([__q_name(i[0]) for i in context.query(
        """SELECT ?i WHERE {
             {%s owl:inverseOf ?i}
             UNION
             {?i owl:inverseOf %s}
           }""" % (prop, prop))])
    return inverses


def get_supertypes(ty, vid=None):
    """

    :param ty:
    :param vid:
    :return:
    """
    context = graph
    if vid is not None:
        context = context.get_context(vid)
    res = map(lambda x: __q_name(x), filter(lambda y: isinstance(y, URIRef),
                                            context.transitive_objects(__extend_prefixed(ty), RDFS.subClassOf)))
    return set(filter(lambda x: str(x) != ty, res))


def get_subtypes(ty, vid=None):
    """

    :param ty:
    :param vid:
    :return:
    """
    context = graph
    if vid is not None:
        context = context.get_context(vid)
    res = map(lambda x: __q_name(x), filter(lambda y: isinstance(y, URIRef),
                                            context.transitive_subjects(RDFS.subClassOf, __extend_prefixed(ty))))

    return filter(lambda x: str(x) != ty, res)


def get_type_properties(ty, vid=None):
    """

    :param ty:
    :param vid:
    :return:
    """
    context = graph
    if vid is not None:
        context = context.get_context(vid)
    props = set([])
    all_types = get_supertypes(ty, vid)
    all_types.add(ty)
    for sc in all_types:
        props.update(
            [__q_name(p[0]) for p in context.query(
                """SELECT ?p WHERE {
                     {%s rdfs:subClassOf [ owl:onProperty ?p ]}
                     UNION
                     {?p rdfs:domain %s}
                     FILTER (isURI(?p))
                   }""" % (sc, sc))])
    return props


def get_type_references(ty, vid=None):
    """

    :param ty:
    :param vid:
    :return:
    """
    context = graph
    if vid is not None:
        context = context.get_context(vid)
    refs = set([])
    all_types = get_supertypes(ty, vid)
    all_types.add(ty)
    for sc in all_types:
        refs.update(
            [__q_name(p[0]) for p in context.query(
                """SELECT ?p WHERE {
                    { ?r owl:onProperty ?p.
                      {?r owl:someValuesFrom %s}
                      UNION
                      {?r owl:allValuesFrom %s}
                      UNION
                      {?r owl:onClass %s}
                    }
                    UNION
                    {?p rdfs:range %s}
                    FILTER (isURI(?p))
                  }""" % (sc, sc, sc, sc))])
    return refs



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

__author__ = 'Fernando Serena'

import base64
import itertools

from agora.fountain.index import core as index
from agora.fountain.index.paths import pgraph
from agora.fountain.index import seeds


def view_graph():
    """

    :return:
    """
    types, nodes, edges, roots = [], [], [], []

    ibase = 0
    nodes_dict = dict([(nid, base64.b16encode(nid)) for nid in pgraph.nodes()])
    for (nid, data) in pgraph.nodes(data=True):
        if data.get('ty') == 'type':
            types.append(nid)
            has_seeds = len(seeds.get_type_seeds(nid))
            node_d = {'data': {'id': nodes_dict[nid], 'label': nid, 'shape': 'roundrectangle',
                               'width': len(nid) * 10}}
            if has_seeds:
                roots.append(nodes_dict[nid])
                node_d['classes'] = 'seed'
            nodes.append(node_d)
            ibase += 1
        elif data.get('ty') == 'prop' and data.get('object'):
            dom = [t for (t, _) in pgraph.in_edges(nid)]
            ran = [t for (_, t) in pgraph.out_edges(nid)]
            dom = [d for d in dom if not set.intersection(set(index.get_type(d).get('super')), set(dom))]
            ran = [r for r in ran if not set.intersection(set(index.get_type(r).get('super')), set(ran))]

            op_edges = list(itertools.product(*[dom, ran]))
            edges.extend([{'data': {'id': 'e{}'.format(ibase + i), 'source': nodes_dict[s], 'label': nid + '\n\n',
                                    'target': nodes_dict[tg]}} for i, (s, tg) in enumerate(op_edges)])
            ibase += len(op_edges) + 1
        else:
            ran = data.get('range')
            if ran is None:
                print nid
            if len(ran):
                dom = [t for (t, _) in pgraph.in_edges(nid)]
                dom = [d for d in dom if not set.intersection(set(index.get_type(d).get('super')), set(dom))]
                dp_edges = list(itertools.product(*[dom, ran]))

                for i, (s, t) in enumerate(dp_edges):
                    rid = 'n{}'.format(len(nodes_dict) + len(nodes))
                    nodes.append({'data': {'id': rid, 'label': t, 'width': len(t) * 10, 'shape': 'ellipse'}})
                    edges.append(
                        {'data': {'id': 'e{}'.format(ibase + i), 'source': nodes_dict[s], 'label': nid + '\n\n',
                                  'target': rid}})
                ibase += len(dp_edges) + 1

    for t in types:
        super_types = index.get_type(t).get('super')
        super_types = [s for s in super_types if not set.intersection(set(index.get_type(s).get('sub')),
                                                                      set(super_types))]
        st_edges = [{'data': {'id': 'e{}'.format(ibase + i), 'source': nodes_dict[st], 'label': '',
                              'target': nodes_dict[t]}, 'classes': 'subclass'}
                    for i, st in enumerate(super_types) if st in nodes_dict]
        if len(st_edges):
            edges.extend(st_edges)

        ibase += len(st_edges) + 1

    return nodes, edges, roots



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
from redis.exceptions import RedisError, BusyLoadingError

__author__ = 'Fernando Serena'

import redis
import agora.fountain.vocab.schema as sch
from agora.fountain.exceptions import FountainError
import agora.fountain.vocab.onto as vocs
from datetime import datetime as dt
from concurrent.futures.thread import ThreadPoolExecutor
import logging
from agora.fountain.server import app
import sys
import multiprocessing

log = logging.getLogger('agora.fountain.index')

redis_conf = app.config['REDIS']
pool = redis.ConnectionPool(host=redis_conf.get('host'), port=redis_conf.get('port'), db=redis_conf.get('db'))
r = redis.StrictRedis(connection_pool=pool)
tpool = ThreadPoolExecutor(multiprocessing.cpu_count())

# Ping redis to check if it's ready
requests = 0
while True:
    try:
        r.keys('*')
        break
    except BusyLoadingError as re:
        log.warning(re.message)
    except RedisError:
        print 'Redis is not available'
        sys.exit(-1)

store_mode = app.config['STORE']
if 'memory' in store_mode:
    r.flushdb()


def __get_by_pattern(pattern, func):
    """

    :param pattern:
    :param func:
    :return:
    """
    def get_all():
        for k in pkeys:
            yield func(k)

    try:
        pkeys = r.keys(pattern)
        return list(get_all())
    except RedisError as e:
        raise FountainError(e.message)


def __remove_from_sets(values, *args):
    """

    :param values:
    :param args:
    :return:
    """
    try:
        for pattern in args:
            keys = r.keys(pattern)
            for dk in keys:
                key_parts = dk.split(':')
                ef_values = values
                if len(key_parts) > 1:
                    ef_values = filter(lambda x: x.split(':')[0] != key_parts[1], values)
                if len(ef_values):
                    r.srem(dk, *ef_values)
    except RedisError as e:
        raise FountainError(e.message)


def __get_vocab_set(pattern, vid=None):
    """

    :param pattern:
    :param vid:
    :return:
    """
    try:
        if vid is not None:
            pattern = pattern.replace(':*:', ':%s:' % vid)
        all_sets = map(lambda x: r.smembers(x), r.keys(pattern))
        return list(reduce(set.union, all_sets, set([])))
    except RedisError as e:
        raise FountainError(e.message)


def __extract_type(t, vid):
    """

    :param t:
    :param vid:
    :return:
    """
    log.debug('Extracting type {} from the {} vocabulary...'.format(t, vid))
    try:
        with r.pipeline() as pipe:
            pipe.multi()
            pipe.sadd('vocabs:{}:types'.format(vid), t)
            for s in sch.get_supertypes(t):
                pipe.sadd('vocabs:{}:types:{}:super'.format(vid, t), s)
            for s in sch.get_subtypes(t):
                pipe.sadd('vocabs:{}:types:{}:sub'.format(vid, t), s)
            for s in sch.get_type_properties(t):
                pipe.sadd('vocabs:{}:types:{}:props'.format(vid, t), s)
            for s in sch.get_type_references(t):
                pipe.sadd('vocabs:{}:types:{}:refs'.format(vid, t), s)
            pipe.execute()
    except RedisError as e:
        raise FountainError(e.message)


def __extract_property(p, vid):
    """

    :param p:
    :param vid:
    :return:
    """
    def p_type():
        if sch.is_object_property(p):
            return 'object'
        else:
            return 'data'

    try:
        log.debug('Extracting property {} from the {} vocabulary...'.format(p, vid))
        with r.pipeline() as pipe:
            pipe.multi()
            pipe.sadd('vocabs:{}:properties'.format(vid), p)
            pipe.hset('vocabs:{}:properties:{}'.format(vid, p), 'uri', p)
            for dc in list(sch.get_property_domain(p)):
                pipe.sadd('vocabs:{}:properties:{}:_domain'.format(vid, p), dc)
            for dc in list(sch.get_property_range(p)):
                pipe.sadd('vocabs:{}:properties:{}:_range'.format(vid, p), dc)
            for dc in list(sch.get_property_inverses(p)):
                pipe.sadd('vocabs:{}:properties:{}:_inverse'.format(vid, p), dc)
            pipe.set('vocabs:{}:properties:{}:_type'.format(vid, p), p_type())
            pipe.execute()
    except RedisError as e:
        raise FountainError(e.message)


def __extract_types(vid):
    """

    :param vid:
    :return:
    """
    types = sch.get_types(vid)

    other_vocabs = filter(lambda x: x != vid, vocs.get_vocabularies())
    dependent_types = set([])
    dependent_props = set([])
    for ovid in other_vocabs:
        o_types = [t for t in get_types(ovid) if t not in types]
        for oty in o_types:
            otype = get_type(oty)
            if set.intersection(types, otype.get('super')) or set.intersection(types, otype.get('sub')):
                dependent_types.add((ovid, oty))
        o_props = [t for t in get_properties(ovid)]
        for op in o_props:
            oprop = get_property(op)
            if set.intersection(types, oprop.get('domain')) or set.intersection(types, oprop.get('range')):
                dependent_props.add((ovid, op))

    types = set.union(set([(vid, t) for t in types]), dependent_types)
    futures = []
    # TODO: Catch exceptions from threadpool
    for v, t in types:
        futures.append(tpool.submit(__extract_type, t, v))
    for v, p in dependent_props:
        futures.append(tpool.submit(__extract_property, p, v))
    return types, futures


def __extract_properties(vid):
    """

    :param vid:
    :return:
    """
    properties = sch.get_properties(vid)

    other_vocabs = filter(lambda x: x != vid, vocs.get_vocabularies())
    dependent_types = set([])
    for ovid in other_vocabs:
        o_types = [t for t in get_types(ovid)]
        for oty in o_types:
            o_type = get_type(oty)
            if set.intersection(properties, o_type.get('refs')) or set.intersection(properties, o_type.get('properties')):
                dependent_types.add((ovid, oty))

    futures = []
    for p in properties:
        futures.append(tpool.submit(__extract_property, p, vid))
    for v, ty in dependent_types:
        futures.append(tpool.submit(__extract_type, ty, v))
    return properties, futures


def delete_vocabulary(vid):
    """

    :param vid:
    :return:
    """
    v_types = get_types(vid)
    if len(v_types):
        __remove_from_sets(v_types, '*:_domain', '*:_range', '*:sub', '*:super')
    v_props = get_properties(vid)
    if len(v_props):
        __remove_from_sets(v_props, '*:refs', '*:props')
    try:
        v_keys = r.keys('vocabs:{}:*'.format(vid))
        if len(v_keys):
            r.delete(*v_keys)
    except RedisError as e:
        raise FountainError(e.message)


def extract_vocabulary(vid):
    """

    :param vid:
    :return:
    """
    log.info('Extracting vocabulary {}...'.format(vid))
    delete_vocabulary(vid)
    start_time = dt.now()
    types, t_futures = __extract_types(vid)
    properties, p_futures = __extract_properties(vid)
    for f in t_futures + p_futures:
        f.result()
    log.info('Done (in {}ms)'.format((dt.now() - start_time).total_seconds() * 1000))
    return types, properties


def get_types(vid=None):
    """

    :param vid:
    :return:
    """
    def shared_type(t):
        return any(filter(lambda k: r.sismember(k, t), all_type_keys))

    vid_types = __get_vocab_set('vocabs:*:types', vid)

    if vid is not None:
        all_type_keys = filter(lambda k: k != 'vocabs:{}:types'.format(vid), r.keys("vocabs:*:types"))
        vid_types = filter(lambda t: not shared_type(t), vid_types)

    return vid_types


def get_properties(vid=None):
    """

    :param vid:
    :return:
    """
    def shared_property(t):
        return any(filter(lambda k: r.sismember(k, t), all_prop_keys))

    vid_props = __get_vocab_set('vocabs:*:properties', vid)

    if vid is not None:
        all_prop_keys = filter(lambda k: k != 'vocabs:{}:properties'.format(vid), r.keys("vocabs:*:properties"))
        vid_props = filter(lambda t: not shared_property(t), vid_props)

    return vid_props


def get_property(prop):
    """

    :param prop:
    :return:
    """
    def get_inverse_domain(ip):
        return reduce(set.union, __get_by_pattern('*:properties:{}:_domain'.format(ip), r.smembers), set([]))

    def get_inverse_range(ip):
        return reduce(set.union, __get_by_pattern('*:properties:{}:_range'.format(ip), r.smembers), set([]))

    all_prop_keys = r.keys('*:properties')
    if not filter(lambda k: r.sismember(k, prop), all_prop_keys):
        raise TypeError('Unknown property')

    domain = reduce(set.union, __get_by_pattern('*:properties:{}:_domain'.format(prop), r.smembers), set([]))
    rang = reduce(set.union, __get_by_pattern('*:properties:{}:_range'.format(prop), r.smembers), set([]))
    inv = reduce(set.union, __get_by_pattern('*:properties:{}:_inverse'.format(prop), r.smembers), set([]))

    if len(inv):
        inverse_dr = [(get_inverse_domain(i), get_inverse_range(i)) for i in inv]
        for dom, ra in inverse_dr:
            domain.update(ra)
            rang.update(dom)

    ty = __get_by_pattern('*:properties:{}:_type'.format(prop), r.get)
    try:
        ty = ty.pop()
    except IndexError:
        ty = 'object'

    return {'domain': list(domain), 'range': list(rang), 'inverse': list(inv), 'type': ty}


def is_property(prop):
    """

    :param prop:
    :return:
    """
    try:
        return len(r.keys('*:properties:{}:*'.format(prop)))
    except RedisError as e:
        raise FountainError(e.message)


def is_type(ty):
    """

    :param ty:
    :return:
    """
    try:
        return len(r.keys('*:types:{}:*'.format(ty)))
    except RedisError as e:
        raise FountainError(e.message)


def get_type(ty):
    """

    :param ty:
    :return:
    """
    try:
        all_type_keys = r.keys('*:types')
        if not filter(lambda k: r.sismember(k, ty), all_type_keys):
            raise TypeError('Unknown type')

        super_types = reduce(set.union, __get_by_pattern('*:types:{}:super'.format(ty), r.smembers), set([]))
        sub_types = reduce(set.union, __get_by_pattern('*:types:{}:sub'.format(ty), r.smembers), set([]))
        type_props = reduce(set.union, __get_by_pattern('*:types:{}:props'.format(ty), r.smembers), set([]))
        type_refs = reduce(set.union, __get_by_pattern('*:types:{}:refs'.format(ty), r.smembers), set([]))

        return {'super': list(super_types),
                'sub': list(sub_types),
                'properties': list(type_props),
                'refs': list(type_refs)}
    except RedisError as e:
        raise FountainError(e.message)



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

__author__ = 'Fernando Serena'


import networkx as nx


def compare_path_graphs(test, pattern):
    """

    :param test:
    :param pattern:
    :return:
    """
    if len(test) == len(pattern):
        for expected in pattern:
            try:
                valid_graph = filter(lambda x: x == expected, test).pop()
                test.remove(valid_graph)
            except IndexError:
                pass
        return not len(test)
    return False


class AgoraGraph(nx.DiGraph):
    def __init__(self, data=None, **attr):
        super(AgoraGraph, self).__init__(data, **attr)

    @property
    def types(self):
        """

        :return:
        """
        return [t for t, data in self.nodes(data=True) if data['ty'] == 'type']

    @property
    def properties(self):
        """

        :return:
        """
        return [t for t, data in self.nodes(data=True) if data['ty'] == 'prop']

    def _check_node_type(self, f, ty):
        """

        :param f:
        :param ty:
        :return:
        """
        def wrapper(name):
            for n, data in self.nodes_iter(data=True):
                if n == name:
                    if data['ty'] == ty:
                        return f(name)
                    raise Exception(ty)
            raise KeyError(ty)

        return wrapper

    def get_type_properties(self, ty):
        """

        :param ty:
        :return:
        """
        return self._check_node_type(self.successors, 'type')(ty)

    def get_type_refs(self, ty):
        """

        :param ty:
        :return:
        """
        return self._check_node_type(self.predecessors, 'type')(ty)

    def get_property_domain(self, prop):
        """

        :param prop:
        :return:
        """
        return self._check_node_type(self.predecessors, 'prop')(prop)

    def get_property_range(self, prop):
        """

        :param prop:
        :return:
        """
        return self._check_node_type(self.successors, 'prop')(prop)

    def get_inverse_property(self, prop):
        """

        :param prop:
        :return:
        """
        cycles = list(nx.simple_cycles(self.copy()))
        for cycle in cycles:
            if prop in cycle and len(cycle) == 4:
                p_index = cycle.index(prop)
                return cycle[p_index - 2]
        return None

    def add_types_from(self, types):
        """

        :param types:
        :return:
        """
        for ty in types:
            self.add_node(ty, ty='type', label=ty)

    def add_type(self, ty):
        """

        :param ty:
        :return:
        """
        self.add_node(ty, ty='type', label=ty)

    def add_properties_from(self, props, obj=True):
        """

        :param props:
        :param obj:
        :return:
        """
        for prop in props:
            self.add_node(prop, ty='prop', object=obj, label=prop)

    def add_property(self, prop, obj=True):
        """

        :param prop:
        :param obj:
        :return:
        """
        self.add_node(prop, ty='prop', object=obj, label=prop)

    def link_types(self, source, link, dest):
        """

        :param source:
        :param link:
        :param dest:
        :return:
        """
        self.add_edges_from([(source, link), (link, dest)])

    def __eq__(self, other):
        """

        :param other:
        :return:
        """
        def match(x, y):
            return x == y

        return nx.is_isomorphic(self, other, node_match=match, edge_match=match)


class PathGraph(AgoraGraph):
    def __init__(self, path=None, cycles=None, data=None, **attr):
        super(PathGraph, self).__init__(data, **attr)
        self.__seeds = set([])
        self.__cycle_graphs = {}
        self.__cycle_ids = []
        self.__prev_type, self.__prev_prop = None, None
        self.root = None
        if type(path) is dict:
            self.__parse_path(path)
        if type(cycles) is list:
            self.__parse_cycles(cycles)

    def __parse_path(self, path):
        """

        :param path:
        :return:
        """
        path_seeds = path['seeds']
        if path_seeds:
            self.__seeds = set(path_seeds)
        raw_steps = path['steps']
        self.__cycle_ids = path.get('cycles', [])

        for step in raw_steps:
            ty, prop = step['type'], step['property']
            self.add_step(ty, prop)

    def __parse_cycles(self, cycles):
        """

        :param cycles:
        :return:
        """
        for cycle in cycles:
            cid = cycle['cycle']
            if cid in self.__cycle_ids:
                cycle_steps = cycle['steps']
                cycle_graph = CycleGraph()
                for step in cycle_steps:
                    ty, prop = step['type'], step['property']
                    cycle_graph.add_step(ty, prop)
                self.__cycle_graphs[cid] = cycle_graph

    def add_step(self, ty, prop):
        """

        :param ty:
        :param prop:
        :return:
        """
        self.add_type(ty)
        self.add_property(prop)
        if self.__prev_prop is not None:
            self.add_edge(self.__prev_prop, ty)
        self.add_edge(ty, prop)
        self.__prev_type, self.__prev_prop = ty, prop
        if self.root is None:
            self.root = ty

    @property
    def seeds(self):
        """

        :return:
        """
        return self.__seeds

    @property
    def cycles(self):
        """

        :return:
        """
        return self.__cycle_graphs.keys()

    def get_cycle(self, cid):
        """

        :param cid:
        :return:
        """
        return self.__cycle_graphs[cid]

    def set_cycle(self, cid, graph):
        """

        :param cid:
        :param graph:
        :return:
        """
        self.__cycle_graphs[cid] = graph

    def __eq__(self, other):
        """

        :param other:
        :return:
        """
        def node_match(a_attr, b_attr):
            return a_attr == b_attr

        def edge_match(a_attr, b_attr):
            return a_attr == b_attr

        def compare_cycles():
            def match_cycle(cid):
                cycle = self.get_cycle(cid)
                for o_cid in other.cycles:
                    o_cycle = other.get_cycle(o_cid)
                    if cycle == o_cycle:
                        return True
                return False

            if len(self.cycles) == len(other.cycles):
                return len(self.cycles) == len(filter(match_cycle, self.cycles))
            return False

        return self.seeds == other.seeds and nx.is_isomorphic(self, other, node_match=node_match,
                                                              edge_match=edge_match) and compare_cycles()


class CycleGraph(PathGraph):
    def __init__(self, path=None, cycles=None, data=None, **attr):
        super(CycleGraph, self).__init__(path=path, cycles=cycles, data=data, **attr)
        self.__cycle_edge_prop = None

    def add_step(self, ty, prop):
        """

        :param ty:
        :param prop:
        :return:
        """
        super(CycleGraph, self).add_step(ty, prop)
        if self.__cycle_edge_prop is not None:
            self.remove_edge(self.__cycle_edge_prop, self.root)
        self.add_edge(prop, self.root)
        self.__cycle_edge_prop = prop


"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

__author__ = 'Fernando Serena'

from nose.tools import *

from agora.fountain.tests import FountainTest

seed_uri = "http://localhost/seed"


class UnknownSeedTest(FountainTest):
    def test_unknown_seed(self):
        seeds = self.seeds
        eq_(len(seeds), False, 'There should not be any seed available')
        self.post_seed("test:Concept1", seed_uri, exp_code=400)


class KnownSeedTest(FountainTest):
    def test_known_seed(self):
        self.post_vocabulary('two_concept_cycle')
        self.post_seed("test:Concept1", seed_uri)
        seeds = self.seeds
        assert 'test:Concept1' in seeds, '%s should be the only seed type'
        c1_seeds = seeds['test:Concept1']
        assert len(c1_seeds) == 1 and seed_uri in c1_seeds, '%s should be the only seed available' % seed_uri


class DuplicateSeedTest(FountainTest):
    def test_duplicate_seed(self):
        self.post_vocabulary('two_concept_cycle')
        self.post_seed("test:Concept1", seed_uri)
        self.post_seed("test:Concept1", seed_uri, exp_code=409)


class ClearSeedTest(FountainTest):
    def test_clear_seed(self):
        self.post_vocabulary('two_concept_cycle')
        self.post_seed("test:Concept1", seed_uri)
        vocabs = self.get_vocabularies()
        self.delete_vocabulary(vocabs[vocabs.keys().pop()])  # It assumes it contains only one
        seeds = self.seeds
        eq_(len(seeds), 0, 'No seed should be kept')


class TwoTypesSeedTest(FountainTest):
    def test_two_types_seeds(self):
        c1 = "test:Concept1"
        c2 = "test:Concept2"
        self.post_vocabulary('two_concept_cycle')
        self.post_seed(c1, seed_uri)
        self.post_seed(c2, seed_uri + '2')
        seeds = self.seeds
        assert c1 in seeds and c2 in seeds, '%s and %s should be the two seed types' % (c1, c2)
        c1_seeds = seeds['test:Concept1']
        assert len(c1_seeds) == 1 and seed_uri in c1_seeds, '%s should be the only seed available of type %s' % (
            seed_uri, c1)
        c2_seeds = seeds['test:Concept2']
        assert len(c2_seeds) == 1 and seed_uri + '2' in c2_seeds, '%s should be the only seed available of type %s' % (
            seed_uri, c2)


class MultipleSeedsOfSameTypeTest(FountainTest):
    def test_multiple_seeds(self):
        c1 = "test:Concept1"
        self.post_vocabulary('two_concept_cycle')
        self.post_seed(c1, seed_uri)
        self.post_seed(c1, seed_uri + '2')
        seeds = self.seeds
        assert c1 in seeds, '%s should be the only seed type' % c1
        c1_seeds = seeds[c1]
        assert len(
            c1_seeds) == 2 and seed_uri in c1_seeds and seed_uri + '2' in c1_seeds, \
            '%s and %s should be the only seeds available of type %s' % (
                seed_uri, seed_uri + '2', c1)


class SeedsByTypeTest(FountainTest):
    def test_type_seeds(self):
        c1 = "test:Concept1"
        self.post_vocabulary('two_concept_cycle')
        self.post_seed(c1, seed_uri)
        c1_seeds = self.get_type_seeds(c1)
        assert len(c1_seeds) == 1 and seed_uri in c1_seeds, '%s should be the only seed available of type %s' % (
            seed_uri, c1)


class SeedsByUnknownTypeTest(FountainTest):
    def test_type_seeds(self):
        c1 = "test:Concept2"
        self.post_vocabulary('two_concept_cycle')
        self.post_seed(c1, seed_uri)
        try:
            self.get_type_seeds('unknown')
            assert False
        except AssertionError:
            pass
